/*
 * tcpWinTMT.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "tcpWinTMT.h"


/*
 * Plugin variables that may be used by other plugins (MUST be declared in
 * the header file as 'extern tcpWin_flow_t *tcpWin_flows;'
 */
tcpWinFlow_t *tcpWinFlows;

static uint32_t pktTcpCnt;  // <-- Aggregated tcp packet counts

static uint32_t winThCntG;  // <-- Aggregated win threshold count
static uint8_t winStatG;    // <-- Aggregated status
static uint8_t winStateG;   // <-- Aggregated state

/*
 * Function prototypes
 */


// Tranalyzer functions

/*
 * This describes the plugin name, version, major and minor version of
 * Tranalyzer required and dependencies
 */
T2_PLUGIN_INIT("tcpWinTMT", "0.8.11", 0, 8);

// * This function is called before processing any packet.
void initialize() {
	// allocate struct for all flows and initialise to 0
	T2_PLUGIN_STRUCT_NEW(tcpWinFlows);

	// output for packet mode
	if (sPktFile) fputs("tcpWinStat\ttcpWinState\ttcpWin\ttcpWinThCnt\tflowTmout\t", sPktFile);

	// register timeouts
	timeout_handler_add(TMT_NEW);
	timeout_handler_add(TMT_BGN);
	timeout_handler_add(TMT_EST);
	timeout_handler_add(TMT_ABRT);
}


/*
 * This function is used to describe the columns output by the plugin
 */
#if BLOCK_BUF == 0
binary_value_t* printHeader() {
	binary_value_t *bv = NULL;

	BV_APPEND_H8(bv, "tcpWinStat", "tcp Windowsize threshold status"); // 8 bit hex variable
	BV_APPEND_H8(bv, "tcpWinState", "tcp internal connection state"); // 8 bit hex variable
        BV_APPEND_U32(bv, "tcpWinThCnt", "tcp Windowsize theshold count"); // 32 unsigned int variable

	return bv;
}
#endif // BLOCK_BUF == 0


/*
 * This function is called every time a new flow is created.
 */
void onFlowGenerated(packet_t *packet __attribute__((unused)), unsigned long flowIndex) {
        flow_t * const flowP = &flows[flowIndex];
	tcpWinFlow_t * const tcpWinFlowP = &tcpWinFlows[flowIndex];
        memset(tcpWinFlowP, '\0', sizeof(*tcpWinFlowP)); // set everything to 0

        if (flowP->status & L2_FLOW) return; // Layer 2 flow. No L3/4 pointers, so return

	flowP->timeout = TMT_NEW;
}


/*
 * This function is called for every packet with a layer 4.
 */
void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
        flow_t * const flowP = &flows[flowIndex];
        if (flowP->layer4Protocol != L3_TCP) return; // process only TCP

        // only 1. frag packet will be processed
        if (!t2_is_first_fragment(packet)) return;

        tcpWinFlow_t *tcpWinFlowP = &tcpWinFlows[flowIndex];

	flow_t *revFlowP;
        tcpWinFlow_t *revtcpWinFlowP;

	if (FLOW_HAS_OPPOSITE(flowP)) {
		revFlowP = &flows[flowP->oppositeFlowIndex];
        	revtcpWinFlowP = &tcpWinFlows[flowP->oppositeFlowIndex];
	} else {
		revFlowP = NULL;
		revtcpWinFlowP = NULL;
	}

	const tcpHeader_t * const tcpHeader = (tcpHeader_t*)packet->layer4Header;
        const uint32_t tcpWin = ntohs(tcpHeader->window);

	pktTcpCnt++; // count all TCP packets

	if (tcpWin < TCPWIN_THRES) {			// is windowsize below threshold?
		tcpWinFlowP->winThCnt++; 		// count the packet
		tcpWinFlowP->stat |= TCPWIN_THU; 	// set the status bit
	}

	const uint8_t tcpFlags = *((uint8_t*)tcpHeader + 13); // copy all tcp flags
	tcpWinFlowP->stat |= tcpFlags & 0x1f; // extract only the main tcp flags (SYN,FIN,RST,ACK,PSH)

	switch (tcpWinFlowP->state) { // internal simplified TCP connection statemachine
		case NEW: // new flow
                        if (tcpFlags & TH_SYN) {
                                flowP->timeout = TMT_BGN;
                                tcpWinFlowP->state = BGN;
                                break;
                        } else if (!(tcpFlags & TH_ACK)) break;
			tcpWinFlowP->state = EST;
		/*FALLTHRU*/
                case BGN: // begin state after 1. packet
                        flowP->timeout = TMT_EST;
		/*FALLTHRU*/
                case EST: // or established connection
                        if (tcpWinFlowP->stat & (TH_FIN | TH_RST)) {
                                tcpWinFlowP->state = ABRT;
                                flowP->timeout = TMT_EST;
                                break;
                        }
        		if (tcpWinFlowP->winThCnt < TCPWIN_FLWPRG) break; // is wincount above threshold?

			// if win count thres met, set abort conditions for flow and rev flow
	                tcpWinFlowP->stat |= TCPWIN_ABNRM; // set abort state
        	        tcpWinFlowP->state = ABRT;	   // set abort status flag
                	flowP->timeout = TMT_ABRT;	   // set abort timeout
			if (revFlowP) {			   // do the same for the reverse flow if it exists
        	        	revtcpWinFlowP->state = ABRT;
				revFlowP->timeout = TMT_ABRT;
			}
		/*FALLTHRU*/
                case ABRT: // abort either FIN or RST
                        flowP->timeout = TMT_ABRT;
                        break;
                default: // anything else do nothing
                        break;

	}

	//output for pkt mode
	if (sPktFile) fprintf(sPktFile, "0x%02"PRIx8"\t0x%02"PRIx8"\t%"PRIu32"\t%"PRIu32"\t%f\t", tcpWinFlowP->stat, tcpWinFlowP->state, tcpWin, tcpWinFlowP->winThCnt, flowP->timeout);
}


void onFlowTerminate(unsigned long flowIndex) {
	tcpWinFlow_t * const tcpWinFlowP = &tcpWinFlows[flowIndex];

	winStatG |= tcpWinFlowP->stat;          // <-- Aggregate all tcp packet flags
	winStateG |= tcpWinFlowP->state;        // <-- Aggregate all states
	winThCntG += tcpWinFlowP->winThCnt;     // <-- Sum all winThCnt packet counts

	OUTBUF_APPEND_U8(main_output_buffer, tcpWinFlowP->stat);
	OUTBUF_APPEND_U8(main_output_buffer, tcpWinFlowP->state);
	OUTBUF_APPEND_U32(main_output_buffer, tcpWinFlowP->winThCnt);
}


// add a endreport
void pluginReport(FILE *stream) {
    if (winThCntG) {
        T2_FPLOG(stream, "tcpWinTMT", "Aggregated status flags: 0x%02"B2T_PRIX8, winStatG);
        T2_FPLOG(stream, "tcpWinTMT", "Aggregated state flags: 0x%02"B2T_PRIX8, winStateG);
        T2_FPLOG(stream, "tcpWinTMT", "Number of tcp winsize packets below threshold %d: %"PRIu32" [%.2f%%]", TCPWIN_THRES, winThCntG, 100.0*(float)winThCntG/pktTcpCnt);
    }
}


/*
 * This function is called once all the packets have been processed.
 * Cleanup all used memory here.
 */
void onApplicationTerminate() {

    free(tcpWinFlows);
}


